/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.apilib.client_attribute.grpc;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.apilib.client_attribute.OzgCloudBooleanClientAttribute;
import de.ozgcloud.apilib.client_attribute.OzgCloudClientAttributeTestFactory;
import de.ozgcloud.apilib.client_attribute.OzgCloudDoubleClientAttribute;
import de.ozgcloud.apilib.client_attribute.OzgCloudLongClientAttribute;
import de.ozgcloud.apilib.client_attribute.OzgCloudStringClientAttribute;
import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextAttachingInterceptor;
import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextProvider;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangTestFactory;
import de.ozgcloud.vorgang.grpc.clientAttribute.ClientAttributeServiceGrpc.ClientAttributeServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcClientAttribute;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcSetClientAttributeRequest;

class OzgCloudClientAttributeGrpcServiceTest {

	@Spy
	@InjectMocks
	private OzgCloudClientAttributeGrpcService service;

	@Mock
	private ClientAttributeServiceBlockingStub serviceBlockingStub;
	@Mock
	private OzgCloudClientAttributeMapper clientAttributeMapper;
	@Mock
	private OzgCloudCallContextProvider contextProvider;
	@Mock
	private GrpcClientAttribute grpcClientAttribute;

	@Nested
	class TestSetClientAttribute {

		@BeforeEach
		void init() {
			doNothing().when(service).doSetClientAttribute(any(GrpcClientAttribute.class), any());
		}

		@Nested
		class TestBoolean {

			private static final OzgCloudBooleanClientAttribute CLIENT_ATTRIBUTE = OzgCloudClientAttributeTestFactory.createBoolean();

			@BeforeEach
			void init() {
				doReturn(grpcClientAttribute).when(clientAttributeMapper).toGrpc(any(OzgCloudBooleanClientAttribute.class));
			}

			@Test
			void shouldCallClientAttributeMapper() {
				service.setClientAttribute(CLIENT_ATTRIBUTE, OzgCloudVorgangTestFactory.ID);

				verify(clientAttributeMapper).toGrpc(CLIENT_ATTRIBUTE);
			}

			@Test
			void shouldCallSetClientAttribute() {
				service.setClientAttribute(CLIENT_ATTRIBUTE, OzgCloudVorgangTestFactory.ID);

				verify(service).doSetClientAttribute(grpcClientAttribute, OzgCloudVorgangTestFactory.ID);
			}
		}

		@Nested
		class TestString {

			private static final OzgCloudStringClientAttribute CLIENT_ATTRIBUTE = OzgCloudClientAttributeTestFactory.createString();

			@BeforeEach
			void init() {
				doReturn(grpcClientAttribute).when(clientAttributeMapper).toGrpc(any(OzgCloudStringClientAttribute.class));
			}

			@Test
			void shouldCallClientAttributeMapper() {
				service.setClientAttribute(CLIENT_ATTRIBUTE, OzgCloudVorgangTestFactory.ID);

				verify(clientAttributeMapper).toGrpc(CLIENT_ATTRIBUTE);
			}

			@Test
			void shouldCallSetClientAttribute() {
				service.setClientAttribute(CLIENT_ATTRIBUTE, OzgCloudVorgangTestFactory.ID);

				verify(service).doSetClientAttribute(grpcClientAttribute, OzgCloudVorgangTestFactory.ID);
			}
		}

		@Nested
		class TestLong {

			private static final OzgCloudLongClientAttribute CLIENT_ATTRIBUTE = OzgCloudClientAttributeTestFactory.createLong();

			@BeforeEach
			void init() {
				doReturn(grpcClientAttribute).when(clientAttributeMapper).toGrpc(any(OzgCloudLongClientAttribute.class));
			}

			@Test
			void shouldCallClientAttributeMapper() {
				service.setClientAttribute(CLIENT_ATTRIBUTE, OzgCloudVorgangTestFactory.ID);

				verify(clientAttributeMapper).toGrpc(CLIENT_ATTRIBUTE);
			}

			@Test
			void shouldCallSetClientAttribute() {
				service.setClientAttribute(CLIENT_ATTRIBUTE, OzgCloudVorgangTestFactory.ID);

				verify(service).doSetClientAttribute(grpcClientAttribute, OzgCloudVorgangTestFactory.ID);
			}
		}

		@Nested
		class TestDouble {

			private static final OzgCloudDoubleClientAttribute CLIENT_ATTRIBUTE = OzgCloudClientAttributeTestFactory.createDouble();

			@BeforeEach
			void init() {
				doReturn(grpcClientAttribute).when(clientAttributeMapper).toGrpc(any(OzgCloudDoubleClientAttribute.class));
			}

			@Test
			void shouldCallClientAttributeMapper() {
				service.setClientAttribute(CLIENT_ATTRIBUTE, OzgCloudVorgangTestFactory.ID);

				verify(clientAttributeMapper).toGrpc(CLIENT_ATTRIBUTE);
			}

			@Test
			void shouldCallSetClientAttribute() {
				service.setClientAttribute(CLIENT_ATTRIBUTE, OzgCloudVorgangTestFactory.ID);

				verify(service).doSetClientAttribute(grpcClientAttribute, OzgCloudVorgangTestFactory.ID);
			}
		}
	}

	@Nested
	class TestDoSetClientAttribute {

		@Mock
		private ClientAttributeServiceBlockingStub stubWithInterceptor;
		@Mock
		private GrpcSetClientAttributeRequest grpcRequest;

		@BeforeEach
		void init() {
			doReturn(stubWithInterceptor).when(service).getAttributesServiceStub();
			doReturn(grpcRequest).when(service).buildRequest(any(), any());
		}

		@Test
		void shouldCallBuildRequest() {
			service.doSetClientAttribute(grpcClientAttribute, OzgCloudVorgangTestFactory.ID);

			verify(service).buildRequest(grpcClientAttribute, OzgCloudVorgangTestFactory.ID.toString());
		}

		@Test
		void shouldCallGetServiceBlockingStub() {
			service.doSetClientAttribute(grpcClientAttribute, OzgCloudVorgangTestFactory.ID);

			verify(service).getAttributesServiceStub();
		}

		@Test
		void shouldCallSet() {
			service.doSetClientAttribute(grpcClientAttribute, OzgCloudVorgangTestFactory.ID);

			verify(stubWithInterceptor).set(grpcRequest);
		}
	}

	@Nested
	class TestBuildRequest {

		private static final String VORGANG_ID = OzgCloudVorgangTestFactory.ID.toString();

		@Test
		void shouldSetClientAttribute() {
			var request = service.buildRequest(grpcClientAttribute, VORGANG_ID);

			assertThat(request.getAttribute()).isSameAs(grpcClientAttribute);
		}

		@Test
		void shouldSetVorgangId() {
			var request = service.buildRequest(grpcClientAttribute, VORGANG_ID);

			assertThat(request.getVorgangId()).isEqualTo(VORGANG_ID);
		}
	}

	@Nested
	class TestGetAttributesServiceStub {

		@Mock
		private ClientAttributeServiceBlockingStub stubWithInterceptor;

		@Captor
		private ArgumentCaptor<OzgCloudCallContextAttachingInterceptor> interceptorCaptor;

		@Test
		void shouldSetContextProvider() {
			service.getAttributesServiceStub();

			verify(serviceBlockingStub).withInterceptors(interceptorCaptor.capture());
			assertThat(getContextProvider()).isSameAs(contextProvider);
		}

		@Test
		void shouldReturnServiceBlockingStub() {
			when(serviceBlockingStub.withInterceptors(any())).thenReturn(stubWithInterceptor);

			var stub = service.getAttributesServiceStub();

			assertThat(stub).isSameAs(stubWithInterceptor);
		}

		private OzgCloudCallContextProvider getContextProvider() {
			return (OzgCloudCallContextProvider) ReflectionTestUtils.getField(interceptorCaptor.getValue(), "callContextProvider");
		}
	}
}